(* ******************************************************************************************


  AMPRNG - I'M PRNG
  Pseudo-random probably safe number generator

  Rev.1.2:

  - More fast AND operations instead of MOD
  - Internal key expansion as setup variant
  - Support for any key length in KSA

  State length: 2048 bit x 2

  Initialization key: 80-512 bit

  Initialization vector: 80-512 bit

  Uses some ideas from RC and VMPC

  Copyright (C) 2009-2010  Alexander Myasnikow

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Affero General Public License for more details.

  You should have received a copy of the GNU Affero General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.


  ******************************************************************************************* *)
{$R-}
{$Q-}
{$D-}
{$O+}
unit amprng;
{$DEFINE PARANOID}

interface

type
  TAMPRNG = class

  var
    P, MG: array [0 .. 255] of byte;
    pos: integer;
    m, z: byte;

  const
    NLBOX: array [0 .. 255] of byte = ($96, $62, $29, $E5, $0C, $B7, $BB, $1E,
      $7C, $1B, $80, $CD, $8D, $3B, $83, $3C, $17, $CB, $11, $87, $16, $1D,
      $C0, $A1, $E9, $DD, $2F, $8F, $89, $FA, $58, $28, $8B, $F2, $3A, $F4,
      $73, $CA, $55, $E7, $94, $9B, $BF, $AF, $1C, $C9, $2B, $74, $BA, $53,
      $98, $75, $BC, $AA, $6C, $EC, $B1, $C4, $08, $01, $C6, $59, $A4, $4A,
      $FE, $C2, $86, $DF, $FF, $56, $6A, $79, $D8, $B8, $B3, $C5, $72, $DC,
      $6D, $32, $69, $4F, $7A, $F9, $7D, $DB, $B4, $92, $9C, $A9, $64, $E1,
      $12, $D6, $FC, $15, $9A, $02, $06, $CF, $19, $D0, $D2, $BE, $B9, $50,
      $47, $23, $5F, $7F, $E6, $5D, $B5, $3E, $48, $33, $EA, $A6, $30, $43,
      $E0, $52, $67, $34, $31, $FD, $00, $AE, $C1, $F6, $18, $27, $97, $F8,
      $E4, $7B, $8C, $2C, $B2, $22, $9F, $10, $AB, $F7, $E2, $5C, $4D, $5B,
      $05, $1A, $39, $03, $0F, $38, $04, $D1, $EE, $60, $99, $41, $E8, $A7,
      $70, $DE, $ED, $81, $7E, $57, $D5, $82, $85, $88, $6B, $0D, $D4, $A3,
      $A8, $EF, $84, $AD, $20, $21, $4C, $35, $66, $42, $91, $0A, $6F, $8E,
      $90, $0B, $F5, $4B, $2D, $D7, $D3, $71, $44, $07, $9D, $FB, $40, $68,
      $BD, $14, $2E, $09, $CC, $1F, $A5, $93, $C8, $C3, $54, $65, $76, $A0,
      $77, $95, $A2, $49, $D9, $AC, $8A, $3F, $24, $36, $45, $F1, $13, $9E,
      $3D, $46, $F0, $B6, $5E, $37, $C7, $6E, $2A, $CE, $63, $26, $E3, $0E,
      $78, $B0, $4E, $5A, $F3, $EB, $51, $61, $25, $DA);

  const
    TKBOX: array [0 .. 255] of byte = ($90, $16, $7C, $A3, $44, $8C, $ED, $83,
      $FC, $65, $C2, $05, $46, $19, $1D, $51, $41, $AA, $4F, $BB, $02, $61,
      $94, $F5, $AC, $C8, $11, $07, $1F, $34, $DE, $D7, $92, $C4, $39, $FB,
      $FA, $E8, $62, $9C, $17, $04, $B6, $DB, $01, $06, $FE, $F2, $B3, $0C,
      $F9, $03, $E3, $EF, $B5, $31, $2C, $B2, $5E, $15, $00, $23, $EE, $53,
      $09, $D6, $3C, $5D, $36, $14, $CF, $72, $6A, $AE, $71, $A6, $56, $84,
      $B7, $AF, $5A, $79, $73, $7D, $2F, $18, $E4, $49, $38, $E1, $08, $47,
      $7A, $3A, $DF, $6C, $91, $66, $8A, $99, $86, $58, $1C, $B4, $43, $8D,
      $BD, $4E, $24, $75, $F8, $E7, $9F, $6B, $A5, $CE, $CB, $CC, $6F, $8E,
      $C6, $80, $54, $9E, $3F, $4A, $0A, $29, $E0, $7E, $45, $BC, $0B, $89,
      $D4, $D9, $95, $D8, $55, $A9, $3D, $5B, $FF, $32, $B8, $8B, $0F, $37,
      $CD, $2B, $57, $69, $78, $A8, $74, $50, $D0, $85, $81, $97, $EA, $4C,
      $6D, $13, $D2, $E2, $70, $10, $F6, $2D, $C1, $D1, $7B, $96, $1B, $26,
      $68, $BA, $B1, $12, $9D, $F0, $AB, $E9, $1E, $BE, $30, $A7, $C3, $8F,
      $F4, $33, $A1, $F1, $20, $F7, $3E, $DA, $0E, $D3, $B0, $42, $64, $67,
      $98, $E5, $D5, $6E, $AD, $E6, $9B, $2E, $88, $CA, $25, $2A, $0D, $4B,
      $52, $93, $1A, $A2, $C7, $C0, $77, $35, $27, $F3, $87, $21, $48, $82,
      $BF, $DC, $B9, $76, $DD, $5C, $60, $59, $40, $22, $EC, $A0, $4D, $28,
      $7F, $EB, $3B, $C9, $9A, $5F, $FD, $63, $C5, $A4);

  const
    MKBOX: array [0 .. 255] of byte = ($CD, $B1, $F3, $3A, $99, $38, $F4, $97,
      $A0, $1A, $62, $C1, $12, $52, $EA, $E8, $F5, $EC, $2C, $09, $7F, $6E,
      $79, $D1, $DF, $88, $75, $BA, $87, $9F, $EE, $53, $AA, $3F, $56, $C2,
      $B5, $48, $D7, $4E, $E4, $03, $33, $F9, $4F, $08, $91, $D6, $1D, $B9,
      $C3, $54, $72, $06, $8B, $0D, $B8, $D8, $9C, $F8, $83, $C8, $4A, $77,
      $FE, $44, $32, $D2, $89, $CC, $20, $43, $9E, $46, $E5, $27, $17, $63,
      $A3, $96, $30, $3D, $DE, $E1, $4C, $29, $C7, $95, $57, $05, $04, $37,
      $AC, $64, $94, $7D, $B7, $59, $82, $D4, $7C, $F6, $CE, $81, $E7, $DC,
      $4B, $8E, $AD, $60, $73, $14, $61, $CB, $66, $A6, $2F, $31, $78, $5C,
      $5B, $41, $C0, $4D, $34, $C4, $F0, $16, $6B, $76, $15, $36, $A9, $71,
      $26, $A4, $E0, $9A, $85, $47, $13, $3C, $11, $58, $2D, $1F, $0C, $7A,
      $D3, $7E, $6F, $6C, $BE, $02, $FD, $3B, $CF, $0F, $A8, $B0, $90, $FA,
      $5D, $DA, $28, $23, $BF, $42, $F1, $6A, $AB, $C9, $AE, $C6, $E6, $D9,
      $1E, $8F, $BD, $50, $86, $E2, $E3, $40, $67, $DD, $AF, $10, $B4, $9B,
      $51, $22, $0B, $5E, $F7, $21, $25, $D5, $2B, $CA, $3E, $65, $1B, $BC,
      $C5, $A7, $FF, $8D, $74, $1C, $49, $92, $B6, $45, $A5, $FC, $98, $00,
      $68, $EF, $D0, $70, $B2, $39, $01, $B3, $2A, $5F, $EB, $5A, $8C, $8A,
      $07, $0E, $F2, $19, $7B, $BB, $18, $80, $84, $24, $A2, $0A, $69, $93,
      $A1, $ED, $9D, $55, $DB, $6D, $2E, $E9, $35, $FB);

  const
    MGBOX: array [0 .. 255] of byte = ($F2, $94, $5A, $4B, $0F, $73, $DA, $DB,
      $83, $1D, $B3, $09, $FC, $36, $B8, $46, $41, $D4, $D0, $55, $F3, $87,
      $76, $9A, $28, $35, $71, $FB, $F7, $1C, $03, $7D, $15, $84, $0A, $43,
      $FA, $9E, $60, $97, $98, $7E, $A3, $52, $D1, $29, $A5, $59, $C5, $7A,
      $2F, $25, $E1, $3B, $38, $0C, $90, $C6, $D9, $F6, $D8, $CF, $16, $95,
      $21, $A7, $6D, $1F, $58, $51, $48, $2A, $BE, $AB, $F1, $5D, $9B, $F9,
      $80, $ED, $E5, $A6, $F5, $6F, $31, $BA, $79, $4F, $85, $81, $B1, $23,
      $E4, $72, $EA, $2C, $CA, $6B, $6A, $1E, $5C, $04, $D5, $AE, $B2, $E6,
      $C3, $39, $4D, $5F, $3A, $45, $B4, $67, $C8, $4E, $1A, $A4, $30, $E0,
      $CC, $10, $BD, $33, $CE, $B7, $E3, $34, $C4, $8A, $FF, $B0, $3F, $02,
      $7C, $DC, $BF, $08, $DF, $8D, $FD, $6C, $EB, $12, $6E, $24, $CD, $2E,
      $99, $5E, $14, $8E, $50, $7F, $AA, $13, $89, $8F, $44, $E7, $88, $B9,
      $20, $26, $A1, $C7, $05, $07, $69, $EF, $DE, $AF, $18, $B6, $DD, $64,
      $01, $D2, $A2, $2B, $0D, $11, $A9, $0B, $BB, $C2, $82, $C1, $C0, $E9,
      $9F, $1B, $EE, $CB, $54, $00, $96, $AD, $27, $74, $5B, $68, $0E, $E8,
      $D6, $22, $A8, $AC, $3E, $3D, $FE, $70, $17, $77, $49, $06, $86, $37,
      $9D, $D7, $53, $63, $3C, $57, $2D, $78, $C9, $75, $8B, $62, $7B, $F8,
      $4C, $F0, $E2, $40, $A0, $93, $B5, $19, $65, $92, $56, $32, $47, $F4,
      $4A, $9C, $8C, $BC, $42, $EC, $D3, $66, $91, $61);

  const
    ESB: array [0 .. 255] of byte = ($74, $65, $2F, $32, $EE, $A2, $FF, $FE,
      $33, $AA, $35, $12, $02, $1A, $56, $A3, $51, $60, $81, $BA, $2B, $09,
      $55, $E9, $EA, $9E, $10, $C1, $4E, $79, $54, $3B, $44, $3D, $C2, $61,
      $66, $8B, $4F, $25, $38, $CA, $E3, $78, $7B, $C9, $00, $47, $FC, $E8,
      $D4, $70, $9D, $2D, $5F, $16, $8E, $03, $62, $EC, $69, $1C, $6F, $C5,
      $AE, $4D, $7A, $8A, $C4, $95, $85, $BF, $23, $DF, $AF, $D5, $22, $28,
      $5C, $B1, $AC, $86, $80, $19, $0D, $14, $B7, $4C, $71, $DB, $FD, $13,
      $7E, $E6, $49, $A4, $D7, $48, $A9, $F5, $0C, $87, $1E, $18, $3C, $F2,
      $CC, $8F, $2C, $57, $68, $D9, $EB, $5E, $F3, $F1, $40, $99, $8C, $2E,
      $83, $36, $53, $6B, $41, $52, $B8, $9A, $DD, $17, $6E, $BB, $4A, $2A,
      $FB, $27, $43, $08, $C6, $A6, $B0, $ED, $CF, $3A, $A8, $42, $64, $9F,
      $82, $30, $31, $3E, $59, $6D, $98, $F6, $BC, $A7, $3F, $04, $B9, $11,
      $0F, $73, $67, $C0, $88, $75, $A5, $0A, $EF, $5D, $C8, $72, $E2, $63,
      $AB, $DA, $D3, $D0, $DE, $E7, $6C, $15, $C7, $D6, $5B, $F8, $45, $91,
      $6A, $A0, $21, $90, $5A, $46, $9B, $DC, $F0, $CD, $07, $B2, $AD, $B4,
      $96, $B6, $97, $76, $D1, $20, $F4, $BD, $D8, $C3, $D2, $7C, $B5, $9C,
      $1F, $1D, $94, $77, $FA, $E4, $BE, $24, $34, $1B, $A1, $01, $29, $58,
      $84, $93, $92, $05, $F9, $E5, $7D, $CE, $7F, $89, $26, $B3, $F7, $4B,
      $37, $CB, $0B, $E0, $8D, $06, $E1, $50, $39, $0E);

    // Normal setup by binary key and IV (optional)
    constructor Create(const key: array of byte; const kl: longint
{$IFDEF PARANOID}; useiv: boolean = False
{$ENDIF}; drop: integer = 0); overload;

    // Easy setup by custom string (password)
    constructor Create(const password: string; const kl: integer;
      drop: integer = 0); overload;

    // IV setup for binary key mode. useiv in Create must be set to true
    procedure SetIV(const iv: array of byte; const ivl: longint);

    function PRNG: byte;
    function RandomU32: longword;
    function Randoms32: longint;
    function RandomU64: uint64;
    function RandomS64: int64;
    procedure GenerateRandom(var buffer: array of byte; len: integer);
    procedure Crypt(var buffer: array of byte; len: integer); overload;
    procedure Crypt(buffer: pointer; len: integer); overload;
    function RandomRange(const AFrom, ATo: integer;
      safe: boolean = False): integer;
    function uRandomRange(const AFrom, ATo: longword;
      safe: boolean = False): longword;
    function Random(const max: integer; safe: boolean = False): integer;
    function uRandom(const max: longword; safe: boolean = False): longword;

  protected
{$IFDEF PARANOID}
    savedkey: array [0 .. 255] of byte;
    savedkl: longint;
    droplen: longint;
{$ENDIF}
    procedure setup(const key: array of byte; const kl: longint; iv: boolean);
    procedure ExpandPassword(material: AnsiString; var key: array of byte;
      kl: integer);
  end;

implementation

type
  TByteArray = array [0 .. 7] of byte;

type
  PByteArray = ^TByteArray;

  {
    State array generation and transformation
    }

procedure TAMPRNG.setup(const key: array of byte; const kl: longint;
  iv: boolean);

var
  i, idx: integer;
  t: byte;
  nl, tk, seed: array [0 .. 255] of byte;

  procedure Morph;
  var
    i: integer;
  begin
    for i := 0 to kl - 1 do
      seed[i] := MKBOX[seed[i]];
  end;

  procedure Cycle;
  var
    i: integer;
  begin
    for i := 0 to 2047 do
    begin
      idx := i and 255;
      m := P[(P[idx] + nl[(m + tk[(seed[i mod kl] + idx) and 255]) and 255])
        and 255];
      t := P[idx];
      P[idx] := P[m];
      P[m] := t;
    end;

  end;

  procedure MGCycle;
  var
    i: integer;
  begin

    for i := 0 to 2047 do
    begin
      idx := i and 255;
      z := MG[(MG[idx] + nl[(z + tk[(seed[i mod kl] + idx) and 255]) and 255])
        and 255];
      t := MG[idx];
      MG[idx] := MG[z];
      MG[z] := t;
    end;

  end;

  procedure SWAP1;
  var
    t: byte;
    i: integer;
  begin
    for i := 0 to 255 do
    begin
      t := P[i];
      P[i] := nl[i];
      nl[i] := t;
    end;
  end;

  procedure SWAP2;
  var
    t: byte;
    i: integer;
  begin
    for i := 0 to 255 do
    begin
      t := P[i];
      P[i] := tk[i];
      tk[i] := t;
    end;

  end;

  procedure SWAP3;
  var
    t: byte;
    i: integer;
  begin
    for i := 0 to 255 do
    begin
      t := nl[i];
      nl[i] := tk[i];
      tk[i] := t;
    end;

  end;

begin

  for i := 0 to kl - 1 do
    seed[i] := key[i];

  for i := 0 to 255 do
  begin
    if not iv then
    begin
      P[i] := i;
    end;
    nl[i] := NLBOX[i];
    tk[i] := TKBOX[i];
  end;

  m := 0;

  Morph;
  Cycle;
  SWAP1;
  Morph;
  Cycle;
  SWAP2;
  Morph;
  Cycle;
  SWAP1;
  Morph;
  Cycle;
  Morph;
  SWAP3;
  Morph;

  if not iv then
  begin
    for i := 0 to 255 do
    begin
      MG[i] := MGBOX[i];
    end;
  end;

  z := 0;
  MGCycle;
  pos := 0;

end;

{
  PRNG init routine:
  generates P and MG states depending on key.

  Set useiv to True if you want to call iv setup routine also.

}

constructor TAMPRNG.Create(const key: array of byte; const kl: integer
{$IFDEF PARANOID}; useiv: boolean = False {$ENDIF}; drop: integer = 0);
var
  i: integer;
begin
  Self.setup(key, kl, False);
{$IFDEF PARANOID}
  if useiv then
  begin

    savedkl := kl;

    for i := 0 to kl - 1 do
    begin
      savedkey[i] := key[i];
    end;

  end;
{$ENDIF}
  droplen := drop;
end;


// No iv in password mode

constructor TAMPRNG.Create(const password: string; const kl: integer;
  drop: integer = 0);
var
  key: array of byte;
begin

  SetLength(key, kl);

  ExpandPassword(password, key, kl);

  Self.setup(key, kl, False);

  droplen := drop;

end;

{
  PRNG init routine:
  transforms P and MG states depending on IV
}

procedure TAMPRNG.SetIV(const iv: array of byte; const ivl: integer);
var
  i: integer;
begin
  Self.setup(iv, ivl, True);
{$IFDEF PARANOID}
  Self.setup(savedkey, savedkl, True);
  for i := 0 to 255 do
  begin
    savedkey[i] := 0;
  end;
  savedkl := 0;
{$ENDIF}
end;

{
  One step of PRNG:
  Transforms P and MG states and outputs one byte
  }

function TAMPRNG.PRNG: byte;
var
  t: byte;
begin

  if droplen = 0 then
    droplen := 1;

  repeat
    m := P[MG[(m + P[pos]) and 255]];
    z := MG[P[(z + MG[pos]) and 255]];
    Result := P[P[(P[m] + 1) and 255]];
    t := P[pos];
    P[pos] := P[m];
    P[m] := t;
    t := MG[pos];
    MG[pos] := MG[z];
    MG[z] := t;
    pos := (pos + 1) and 255;
    Dec(droplen);
  until droplen = 0;

end;

{

  Example routines - PRNG wrappers

}

{
  Random 32-bit unsigned integer
}

function TAMPRNG.RandomU32;
begin
  Result := 0;
  PByteArray(@Result)^[0] := PRNG;
  PByteArray(@Result)^[1] := PRNG;
  PByteArray(@Result)^[2] := PRNG;
  PByteArray(@Result)^[3] := PRNG;
end;

{
  Random 32-bit signed integer
}

function TAMPRNG.Randoms32;
begin
  Result := 0;
  PByteArray(@Result)^[0] := PRNG;
  PByteArray(@Result)^[1] := PRNG;
  PByteArray(@Result)^[2] := PRNG;
  PByteArray(@Result)^[3] := PRNG;
end;

{
  Random 64-bit unsigned integer
}

function TAMPRNG.RandomU64;
begin
  Result := 0;
  PByteArray(@Result)^[0] := PRNG;
  PByteArray(@Result)^[1] := PRNG;
  PByteArray(@Result)^[2] := PRNG;
  PByteArray(@Result)^[3] := PRNG;
  PByteArray(@Result)^[4] := PRNG;
  PByteArray(@Result)^[5] := PRNG;
  PByteArray(@Result)^[6] := PRNG;
  PByteArray(@Result)^[7] := PRNG;
end;

{
  Random 64-bit signed integer
}

function TAMPRNG.RandomS64;
begin
  Result := 0;
  PByteArray(@Result)^[0] := PRNG;
  PByteArray(@Result)^[1] := PRNG;
  PByteArray(@Result)^[2] := PRNG;
  PByteArray(@Result)^[3] := PRNG;
  PByteArray(@Result)^[4] := PRNG;
  PByteArray(@Result)^[5] := PRNG;
  PByteArray(@Result)^[6] := PRNG;
  PByteArray(@Result)^[7] := PRNG;
end;

{
  Genetates array of random bytes
}

procedure TAMPRNG.GenerateRandom(var buffer: array of byte; len: integer);
var
  i: integer;
begin
  for i := 0 to len - 1 do
    buffer[i] := PRNG;

end;

{
  Encrypts array of bytes
}

procedure TAMPRNG.Crypt(var buffer: array of byte; len: integer);
var
  i: integer;
begin
  for i := 0 to len - 1 do
    buffer[i] := buffer[i] xor PRNG;

end;

procedure TAMPRNG.Crypt(buffer: pointer; len: integer);
var
  i: integer;
begin
  for i := 0 to len - 1 do
  begin
    byte(buffer^) := byte(buffer^) xor PRNG;
    longword(buffer) := longword(buffer) + 1;
  end;

end;

function TAMPRNG.RandomRange(const AFrom, ATo: integer;
  safe: boolean = False): integer;
begin

  if AFrom > ATo then
    Result := Self.Random(AFrom - ATo, safe) + ATo
  else
    Result := Self.Random(ATo - AFrom, safe) + AFrom;

end;

function TAMPRNG.uRandomRange(const AFrom, ATo: longword;
  safe: boolean = False): longword;
begin

  if AFrom > ATo then
    Result := Self.uRandom(AFrom - ATo, safe) + ATo
  else
    Result := Self.uRandom(ATo - AFrom, safe) + AFrom;

end;

function TAMPRNG.Random(const max: integer; safe: boolean = False): integer;
var
  tmp: longword;
begin

  Result := 0;

  case safe of
    False:
      Result := Self.RandomU32 mod max;
    True:
      begin
        repeat
          tmp := Self.RandomU32;
        until tmp < max;

        Result := tmp;
      end;

  end;

end;

function TAMPRNG.uRandom(const max: longword; safe: boolean = False): longword;
begin

  Result := 0;

  case safe of
    False:
      Result := Self.RandomU32 mod max;
    True:
      repeat
        Result := Self.RandomU32;
      until Result < max;

    end;

  end;


  // For first user password string to binary key expansion

  procedure TAMPRNG.ExpandPassword(material: AnsiString;
    var key: array of byte; kl: integer);
  var
    keymat: array [0 .. 63] of byte;
    i, L, max, step: integer;

    function rotate(a, b: byte): byte;
  asm
    MOV     AL, a
    MOV     CL, b
    ROL     AL, CL
  end;

    begin

      L := length(material);

      for i := 0 to 63 do
        keymat[i] := 0;

      if L > 63 then
        max := L - 1
      else
        max := 63;

      step := 0;

      for i := 0 to max do
      begin

        if i and 7 = 7 then
          step := step + 1;

        if i < (L - 1) then
          keymat[i and 63] := ESB[keymat[i and 63] + byte
            (material[(i mod L) + 1])]
        else
          keymat[i and 63] := ESB[keymat[i and 63] + rotate
            (byte(material[(i mod L) + 1]), ((i + step) mod 7) + 1)];

      end;

      setup(keymat, 64, False);

      droplen := 768;

      GenerateRandom(key, kl);

    end;

  (* ******************************************************************************************

    Usage:

    Call create(key, key_length)
    Call setiv (iv, iv_length)

    or

    Call create(password, key_length)

    for n = 1 to random_length do random[n] := prng

    This gives you random bytes

    if you want to crypt some data
    use this:

    Call create(key, key_length)
    Call setiv (iv, iv_length)

    for n = 1 to data_length do data[n] = data[n] xor prng

    ******************************************************************************************* *)

end.
